// SPDX-License-Identifier: GPL-2.0
/*
 * Copyright (c) 2000-2005 Silicon Graphics, Inc.
 * All Rights Reserved.
 */
#ifndef __LIBFROG_UTIL_H__
#define __LIBFROG_UTIL_H__

unsigned int	log2_roundup(unsigned int i);

#endif /* __LIBFROG_UTIL_H__ */
